from __future__ import annotations

import logging
import math

from collections import deque
from dataclasses import dataclass, field
from typing import Literal

logger = logging.getLogger(__name__)


@dataclass
class DBSCAN:
    """
    DBSCAN algorithm

    Examples:
        >>> from toyml.clustering import DBSCAN
        >>> dataset = [[1, 2], [2, 2], [2, 3], [8, 7], [8, 8], [25, 80]]
        >>> dbscan = DBSCAN(eps=3, min_samples=2).fit(dataset)
        >>> dbscan.clusters_
        [[0, 1, 2], [3, 4]]
        >>> dbscan.noises_
        [5]
        >>> dbscan.labels_
        [0, 0, 0, 1, 1, -1]

    Tip: References
        1. Zhou Zhihua
        2. Han
        3. Kassambara
    """

    eps: float = 0.5
    """The maximum distance between two samples for one to be considered as in the neighborhood of the other.
    This is not a maximum bound on the distances of points within a cluster.
    This is the most important DBSCAN parameter to choose appropriately for your data set and distance function.
    (same as sklearn)
    """
    min_samples: int = 5
    """The number of samples (or total weight) in a neighborhood for a point to be considered as a core point.
    This includes the point itself. If min_samples is set to a higher value,
    DBSCAN will find denser clusters, whereas if it is set to a lower value, the found clusters will be more sparse.
    (same as sklearn)
    """
    clusters_: list[list[int]] = field(default_factory=list)
    """The clusters found by the DBSCAN algorithm."""
    core_objects_: set[int] = field(default_factory=set)
    """The core objects found by the DBSCAN algorithm."""
    noises_: list[int] = field(default_factory=list)
    """The noises found by the DBSCAN algorithm."""
    labels_: list[int] = field(default_factory=list)
    """The cluster labels found by the DBSCAN algorithm."""

    def fit(self, data: list[list[float]]) -> "DBSCAN":
        """
        Fit the DBSCAN model.

        Args:
            data: The dataset.

        Returns:
            self: The fitted DBSCAN model.
        """
        dataset = Dataset(data)

        # initialize the unvisited set
        unvisited = set(range(dataset.n))
        # get core objects
        self.core_objects_, self.noises_ = dataset.get_core_objects(self.eps, self.min_samples)

        # core objects used for training
        if len(self.core_objects_) == 0:
            logger.warning("No core objects found, all data points are noise. Try to adjust the hyperparameters.")
            return self

        # set of core objects: unordered
        core_object_set = self.core_objects_.copy()
        while core_object_set:
            unvisited_old = unvisited.copy()
            core_object = core_object_set.pop()
            queue: deque[int] = deque()
            queue.append(core_object)
            unvisited.remove(core_object)

            while queue:
                q = queue.popleft()
                neighbors = dataset.get_neighbors(q, self.eps)
                if len(neighbors) + 1 >= self.min_samples:
                    delta = set(neighbors) & unvisited
                    for point in delta:
                        queue.append(point)
                        unvisited.remove(point)

            cluster = unvisited_old - unvisited
            self.clusters_.append(list(cluster))
            core_object_set -= cluster

        self.labels_ = [-1] * dataset.n  # -1 means noise
        for i, cluster_index in enumerate(self.clusters_):
            for j in cluster_index:
                self.labels_[j] = i
        return self

    def fit_predict(self, data: list[list[float]]) -> list[int]:
        """
        Fit the DBSCAN model and return the cluster labels.

        Args:
            data: The dataset.

        Returns:
            The cluster labels.
        """
        return self.fit(data).labels_


@dataclass
class Dataset:
    """
    Dataset for DBSCAN

    Args:
        data: The dataset.

    Attributes:
        data: The dataset.
        n: The number of data points.
        distance_matrix_: The distance matrix.
    """

    data: list[list[float]]
    n: int = field(init=False)
    distance_metric: Literal["euclidean"] = "euclidean"
    """The distance metric to use.(For now we only support euclidean)."""
    distance_matrix_: list[list[float]] = field(init=False)

    def __post_init__(self) -> None:
        self.n = len(self.data)
        self.distance_matrix_ = self._calculate_distance_matrix()

    def _calculate_distance_matrix(self) -> list[list[float]]:
        dist_mat = [[0.0 for _ in range(self.n)] for _ in range(self.n)]
        for i in range(self.n):
            for j in range(i, self.n):
                dist_mat[i][j] = self._get_distance(self.data[i], self.data[j])
                dist_mat[j][i] = dist_mat[i][j]
        return dist_mat

    def get_neighbors(self, i: int, eps: float) -> list[int]:
        """
        Get the neighbors of the i-th data point.

        Args:
            i: The index of the data point.
            eps: The maximum distance between two samples for one to be considered as in the neighborhood of the other.

        Returns:
            The indices of the neighbors (Don't include the point itself).
        """
        return [j for j in range(self.n) if i != j and self.distance_matrix_[i][j] <= eps]

    def get_core_objects(self, eps: float, min_samples: int) -> tuple[set[int], list[int]]:
        """
        Get the core objects and noises of the dataset.

        Args:
            eps: The maximum distance between two samples for one to be considered as in the neighborhood of the other.
            min_samples: The number of samples (or total weight) in a neighborhood for a point to be considered as a core point.

        Returns:
            core_objects: The indices of the core objects.
            noises: The indices of the noises.
        """
        core_objects = set()
        noises = []
        for i in range(self.n):
            neighbors = self.get_neighbors(i, eps)
            if len(neighbors) + 1 >= min_samples:  # +1 to include the point itself
                core_objects.add(i)
            else:
                noises.append(i)
        return core_objects, noises

    def _get_distance(self, x: list[float], y: list[float]) -> float:
        assert len(x) == len(y), f"{x} and {y} have different length!"
        if self.distance_metric == "euclidean":
            return math.sqrt(sum(pow(x[i] - y[i], 2) for i in range(len(x))))
        else:
            raise ValueError(f"Distance metric {self.distance_metric} not supported!")


if __name__ == "__main__":
    dataset: list[list[float]] = [[1.0, 2], [2, 2], [2, 3], [8, 7], [8, 8], [25, 80]]
    dbscan = DBSCAN(eps=3, min_samples=2).fit(dataset)
    for i, cluster in enumerate(dbscan.clusters_):
        print(f"cluster {i}: {[dataset[i] for i in cluster]}")
    print("noise: ", [dataset[i] for i in dbscan.noises_])



from __future__ import annotations

import math
import random

from dataclasses import dataclass, field
from typing import Optional


def bst_expect_length(n: int) -> float:
    if n <= 1:
        return 0
    if n == 2:
        return 1
    return 2 * (math.log(n - 1) + 0.5772156649) - (2 * (n - 1) / n)


@dataclass
class IsolationTree:
    """
    The isolation tree.

    Note:
        The isolation tree is a proper(full) binary tree, which has either 0 or 2 children.
    """

    max_height: int
    """The maximum height of the tree."""
    random_seed: int | None = None
    """The random seed used to initialize the centroids."""
    sample_size_: int | None = None
    """The sample size."""
    feature_num_: int | None = None
    """The number of features at each sample."""
    left_: IsolationTree | None = None
    """The left child of the tree."""
    right_: IsolationTree | None = None
    """The right child of the tree."""
    split_at_: int | None = None
    """The index of feature which is used to split the tree's samples into children."""
    split_value_: float | None = None
    """The value of split_at feature that use to split samples"""

    def __post_init__(self) -> None:
        self.random_state = random.Random(self.random_seed)
        if self.max_height < 0:
            raise ValueError(f"The max height of {self.__class__.__name__} must >= 0, not get {self.max_height}")

    def fit(self, samples: list[list[float]]) -> IsolationTree:
        """
        Fit the isolation tree.
        """
        self.sample_size_ = len(samples)
        self.feature_num_ = len(samples[0])
        # exNode
        if self.max_height == 0 or self.sample_size_ == 1:
            return self
        # inNode
        left_itree, right_itree = self._get_left_right_child_itree(samples)
        self.left_, self.right_ = left_itree, right_itree
        return self

    def get_sample_path_length(self, sample: list[float]) -> float:
        """
        Get the sample's path length to the external(leaf) node.

        Args:
            sample: The data sample.

        Returns:
            The path length of the sample.
        """
        if self.is_external_node():
            assert self.sample_size_ is not None
            if self.sample_size_ == 1:
                return 0
            else:
                return bst_expect_length(self.sample_size_)

        assert self.split_at_ is not None and self.split_value_ is not None
        if sample[self.split_at_] < self.split_value_:
            assert self.left_ is not None
            return 1 + self.left_.get_sample_path_length(sample)
        else:
            assert self.right_ is not None
            return 1 + self.right_.get_sample_path_length(sample)

    def is_external_node(self) -> bool:
        """
        The tree node is external(leaf) node or not.
        """
        if self.left_ is None and self.right_ is None:
            return True
        return False

    def _get_left_right_child_itree(
        self, samples: list[list[float]]
    ) -> tuple[Optional[IsolationTree], Optional[IsolationTree]]:
        assert self.feature_num_ is not None
        split_at_list = list(range(self.feature_num_))
        self.random_state.shuffle(split_at_list)
        for split_at in split_at_list:
            split_at_feature_values = [sample[split_at] for sample in samples]
            split_at_min, split_at_max = min(split_at_feature_values), max(split_at_feature_values)
            if math.isclose(split_at_min, split_at_max):
                continue
            split_value = self.random_state.uniform(split_at_min, split_at_max)
            left_samples, right_samples = self._get_sub_samples_by_split(samples, split_at, split_value)
            # need to keep proper binary tree property: all internal nodes have exactly two children
            if len(left_samples) > 0 and len(right_samples) > 0:
                self.split_at_, self.split_value_ = split_at, split_value
                left_itree = IsolationTree(max_height=self.max_height - 1).fit(left_samples)
                right_itree = IsolationTree(max_height=self.max_height - 1).fit(right_samples)
                return left_itree, right_itree
        # cannot split the samples by any features
        return None, None

    @staticmethod
    def _get_sub_samples_by_split(
        samples: list[list[float]],
        split_at: int,
        split_value: float,
    ) -> tuple[list[list[float]], list[list[float]]]:
        left_samples, right_samples = [], []
        for sample in samples:
            if sample[split_at] < split_value:
                left_samples.append(sample)
            else:
                right_samples.append(sample)
        return left_samples, right_samples


@dataclass
class IsolationForest:
    """
    Isolation Forest.

    Examples:
        >>> from toyml.ensemble.iforest import IsolationForest
        >>> dataset = [[-1.1], [0.3], [0.5], [100.0]]
        >>> IsolationForest(n_itree=100, max_samples=4).fit_predict(dataset)
        [1, 1, 1, -1]

    References:
        Liu, Fei Tony, Kai Ming Ting, and Zhi-Hua Zhou. "Isolation forest." 2008 eighth ieee international conference on data mining. IEEE, 2008.
    """

    n_itree: int = 100
    """The number of isolation tree in the ensemble."""
    max_samples: None | int = None
    """The number of samples to draw from X to train each base estimator."""
    score_threshold: float = 0.5
    """The score threshold that is used to define outlier:
    If sample's anomaly score > score_threshold, then the sample is detected as outlier(predict return -1);
    otherwise, the sample is normal(predict return 1)"""
    random_seed: int | None = None
    """The random seed used to initialize the centroids."""
    itrees_: list[IsolationTree] = field(default_factory=list)
    """The isolation trees in the forest."""

    def __post_init__(self) -> None:
        self.random_state = random.Random(self.random_seed)

    def fit(self, dataset: list[list[float]]) -> IsolationForest:
        """
        Fit the isolation forest model.
        """
        if self.max_samples is None or self.max_samples > len(dataset):
            self.max_samples = len(dataset)

        self.itrees_ = self._fit_itrees(dataset)
        return self

    def score(self, sample: list[float]) -> float:
        """
        Predict the sample's anomaly score.

        Args:
            sample: The data sample.

        Returns:
            The anomaly score.
        """
        assert len(self.itrees_) == self.n_itree, "Please fit the model before score sample!"
        assert self.max_samples is not None, "Please fit the model before score sample!"
        itree_path_lengths = [itree.get_sample_path_length(sample) for itree in self.itrees_]
        expect_path_length = sum(itree_path_lengths) / len(itree_path_lengths)
        score = 2 ** (-expect_path_length / bst_expect_length(self.max_samples))
        return score

    def predict(self, sample: list[float]) -> int:
        """
        Predict the sample is outlier ot not.

        Args:
            sample: The data sample.

        Returns:
            Outlier: -1; Normal: 1.
        """
        score = self.score(sample)
        # outlier
        if score > self.score_threshold:
            return -1
        else:
            return 1

    def fit_predict(self, dataset: list[list[float]]) -> list[int]:
        self.fit(dataset)
        return [self.predict(sample) for sample in dataset]

    def _fit_itrees(self, dataset: list[list[float]]) -> list[IsolationTree]:
        itrees = [self._fit_itree(dataset) for _ in range(self.n_itree)]
        return itrees

    def _fit_itree(self, dataset: list[list[float]]) -> IsolationTree:
        assert self.max_samples is not None, "Please fit the model before score sample!"
        samples = self.random_state.sample(dataset, self.max_samples)
        itree_max_height = math.ceil(math.log2(len(samples)))
        return IsolationTree(max_height=itree_max_height, random_seed=self.random_seed).fit(samples)


if __name__ == "__main__":
    simple_dataset = [[-1.1], [0.3], [0.5], [100.0]]
    result = IsolationForest(random_seed=42).fit_predict(simple_dataset)
    print(result)



from .agnes import AGNES
from .bisect_kmeans import BisectingKmeans
from .dbscan import DBSCAN
from .kmeans import Kmeans

__all__ = [
    "Kmeans",
    "BisectingKmeans",
    "DBSCAN",
    "AGNES",
]



from __future__ import annotations

import math
import statistics

from dataclasses import dataclass, field
from typing import Optional

from toyml.clustering.kmeans import Kmeans


@dataclass
class ClusterTree:
    """
    Cluster tree node.
    The cluster tree is a binary tree, each node is a cluster.
    The root node is the whole dataset, and each node is split into two clusters until the number of clusters is equal to the specified number of clusters.
    The cluster is represented by the indices of the dataset.
    The cluster tree is used to store the cluster information and the relationship between clusters.
    """

    parent: Optional[ClusterTree] = None
    """Parent node."""
    left: Optional[ClusterTree] = None
    """Left child node."""
    right: Optional[ClusterTree] = None
    """Right child node."""
    cluster: list[int] = field(default_factory=list)
    """The cluster: dataset sample indices."""
    centroid: Optional[list[float]] = None
    """The centroid of the cluster."""

    def is_root(self) -> bool:
        return self.parent is None

    def is_leaf(self) -> bool:
        return self.left is None and self.right is None

    def leaf_cluster_nodes(self) -> list[ClusterTree]:
        """Get all the leaves in the cluster tree, which are the clusters of dataset"""
        clusters = []

        def dfs(node: ClusterTree) -> None:
            # only collect the leaf nodes
            if node.is_leaf():
                clusters.append(node)
            if node.left:
                dfs(node.left)
            if node.right:
                dfs(node.right)

        dfs(self)
        return clusters

    def get_clusters(self) -> list[list[int]]:
        """
        Get all clusters(cluster in leaf nodes).
        """
        clusters = [cluster_node.cluster for cluster_node in self.leaf_cluster_nodes()]
        return clusters

    def add_left_child(self, node: ClusterTree) -> ClusterTree:
        self.left = node
        node.parent = self
        return self

    def add_right_child(self, node: ClusterTree) -> ClusterTree:
        self.right = node
        node.parent = self
        return self

    def plot(self) -> None:  # pragma: no cover
        """
        Plot the cluster tree with adaptive node sizes.
        """
        import matplotlib.pyplot as plt
        import networkx as nx

        def _build_graph(
            node: ClusterTree,
            graph: Optional[nx.Graph] = None,
            pos: Optional[dict[int, tuple[float, float]]] = None,
            x: float = 0,
            y: float = 0,
            layer: int = 1,
        ) -> tuple[nx.Graph, dict[int, tuple[float, float]]]:
            if graph is None:
                graph = nx.Graph()
                pos = {}

            node_id = id(node)
            graph.add_node(node_id)
            pos[node_id] = (x, y)  # type: ignore
            graph.nodes[node_id]["label"] = f"{node.cluster}"
            graph.nodes[node_id]["size"] = len(node.cluster) * 1000  # Adjust node size based on cluster size

            if node.left:
                left_id = id(node.left)
                graph.add_edge(node_id, left_id)
                l_x, l_y = x - 1 / 2 ** (layer + 1), y - 0.5
                _build_graph(node.left, graph, pos, l_x, l_y, layer + 1)

            if node.right:
                right_id = id(node.right)
                graph.add_edge(node_id, right_id)
                r_x, r_y = x + 1 / 2 ** (layer + 1), y - 0.5
                _build_graph(node.right, graph, pos, r_x, r_y, layer + 1)

            return graph, pos  # type: ignore

        G, pos = _build_graph(self)

        plt.figure(figsize=(12, 8))

        # Get node sizes
        node_sizes = [G.nodes[node]["size"] for node in G.nodes()]

        nx.draw(G, pos, with_labels=False, node_color="lightblue", node_size=node_sizes, arrows=False)

        labels = nx.get_node_attributes(G, "label")
        nx.draw_networkx_labels(G, pos, labels, font_size=8)  # Reduced font size for better fit

        plt.axis("off")
        plt.title("Cluster Tree")
        # plt.tight_layout()
        plt.savefig("cluster_tree.png", dpi=300, bbox_inches="tight")
        plt.show()


@dataclass
class BisectingKmeans:
    """
    Bisecting K-means algorithm.
    Belong to Divisive hierarchical clustering (DIANA) algorithm.(top-down)

    Examples:
        >>> from toyml.clustering import BisectingKmeans
        >>> dataset = [[1.0, 1.0], [1.0, 2.0], [2.0, 1.0], [10.0, 1.0], [10.0, 2.0], [11.0, 1.0]]
        >>> bisect_kmeans = BisectingKmeans(k=3)
        >>> labels = bisect_kmeans.fit_predict(dataset)
        >>> clusters = bisect_kmeans.cluster_tree_.get_clusters()

    References:
        1. Harrington
        2. Tan

    See Also:
      - K-means algorithm: [toyml.clustering.kmeans][]
    """

    k: int
    """The number of clusters, specified by user."""
    cluster_tree_: ClusterTree = field(default_factory=ClusterTree)
    """The cluster tree"""
    labels_: list[int] = field(default_factory=list)
    """The cluster labels of the dataset."""

    def fit(self, dataset: list[list[float]]) -> "BisectingKmeans":
        """
        Fit the dataset with Bisecting K-means algorithm.

        Args:
            dataset: The set of data points for clustering.

        Returns:
            self.

        """
        n = len(dataset)
        # check dataset
        if self.k > n:
            raise ValueError(
                f"Number of clusters(k) cannot be greater than the number of samples(n), not get {self.k=} > {n=}"
            )
        # start with only one cluster which contains all the data points in dataset
        cluster = list(range(n))
        self.cluster_tree_.cluster = cluster
        self.cluster_tree_.centroid = self._get_cluster_centroids(dataset, cluster)
        self.labels_ = self._predict_dataset_labels(dataset)
        total_error = sum_square_error([dataset[i] for i in cluster])
        # iterate until got k clusters
        while len(self.cluster_tree_.get_clusters()) < self.k:
            # init values for later iteration
            to_splot_cluster_node = None
            split_cluster_into: Optional[tuple[list[int], list[int]]] = None
            for cluster_index, cluster_node in enumerate(self.cluster_tree_.leaf_cluster_nodes()):
                # perform K-means with k=2
                cluster_data = [dataset[i] for i in cluster_node.cluster]
                # If the cluster cannot be split further, skip it
                if len(cluster_data) < 2:
                    continue
                # Bisect by kmeans with k=2
                cluster_unsplit_error, cluster_split_error, (cluster1, cluster2) = self._bisect_by_kmeans(
                    cluster_data, cluster_node, dataset
                )
                new_total_error = total_error - cluster_unsplit_error + cluster_split_error
                if new_total_error < total_error:
                    total_error = new_total_error
                    split_cluster_into = (cluster1, cluster2)
                    to_splot_cluster_node = cluster_node

            if to_splot_cluster_node is not None and split_cluster_into is not None:
                self._commit_split(to_splot_cluster_node, split_cluster_into, dataset)
                self.labels_ = self._predict_dataset_labels(dataset)
            else:
                break

        return self

    def fit_predict(self, dataset: list[list[float]]) -> list[int]:
        """
        Fit and predict the cluster label of the dataset.

        Args:
            dataset: The set of data points for clustering.

        Returns:
            Cluster labels of the dataset samples.
        """
        return self.fit(dataset).labels_

    def predict(self, points: list[list[float]]) -> list[int]:
        """
        Predict the cluster label of the given points.

        Args:
            points: A list of data points to predict.

        Returns:
            A list of predicted cluster labels for the input points.

        Raises:
            ValueError: If the model has not been fitted yet.
        """
        if self.cluster_tree_.centroid is None:
            raise ValueError("The model has not been fitted yet.")

        clusters = self.cluster_tree_.get_clusters()
        predictions = []
        for point in points:
            node = self.cluster_tree_
            while not node.is_leaf():
                if node.left is None or node.right is None:
                    raise ValueError("Invalid cluster tree structure.")

                dist_left = euclidean_distance(point, node.left.centroid)  # type: ignore[arg-type]
                dist_right = euclidean_distance(point, node.right.centroid)  # type: ignore[arg-type]

                node = node.left if dist_left < dist_right else node.right
            cluster_index = clusters.index(node.cluster)
            predictions.append(cluster_index)

        return predictions

    @staticmethod
    def _bisect_by_kmeans(
        cluster_data: list[list[float]],
        cluster_node: ClusterTree,
        dataset: list[list[float]],
    ) -> tuple[float, float, tuple[list[int], list[int]]]:
        kmeans = Kmeans(k=2).fit(cluster_data)
        assert kmeans.clusters_ is not None
        cluster1, cluster2 = kmeans.clusters_[0], kmeans.clusters_[1]
        # Note: map the cluster's inner index to the truth index in dataset
        cluster1 = [cluster_node.cluster[i] for i in cluster1]
        cluster2 = [cluster_node.cluster[i] for i in cluster2]
        # split error calculation
        cluster_unsplit_error = sum_square_error([dataset[i] for i in cluster_node.cluster])
        cluster_split_error = sum_square_error([dataset[i] for i in cluster1]) + sum_square_error(
            [dataset[i] for i in cluster2]
        )
        return cluster_unsplit_error, cluster_split_error, (cluster1, cluster2)

    def _commit_split(
        self,
        cluster_node: ClusterTree,
        split_cluster_into: tuple[list[int], list[int]],
        dataset: list[list[float]],
    ) -> None:
        # cluster tree
        cluster_node.add_left_child(
            ClusterTree(
                cluster=split_cluster_into[0], centroid=self._get_cluster_centroids(dataset, split_cluster_into[0])
            )
        )
        cluster_node.add_right_child(
            ClusterTree(
                cluster=split_cluster_into[1], centroid=self._get_cluster_centroids(dataset, split_cluster_into[1])
            )
        )

    def _predict_dataset_labels(self, dataset: list[list[float]]) -> list[int]:
        labels = [-1] * len(dataset)
        for cluster_label, cluster in enumerate(self.cluster_tree_.get_clusters()):
            for data_point_index in cluster:
                labels[data_point_index] = cluster_label
        return labels

    @staticmethod
    def _get_cluster_centroids(dataset: list[list[float]], cluster: list[int]) -> list[float]:
        cluster_points = [dataset[i] for i in cluster]
        centroid = [sum(t) / len(cluster) for t in zip(*cluster_points)]
        return centroid


def euclidean_distance(v1: list[float], v2: list[float]) -> float:
    """
    Calculate the L2 distance between two vectors

    """
    assert len(v1) == len(v2), f"{v1} and {v2} have different length!"
    return math.sqrt(sum(pow(v1[i] - v2[i], 2) for i in range(len(v1))))


def sum_square_error(c: list[list[float]]) -> float:
    """
    Calc the sum of squared errors.
    """
    mean_c = [statistics.mean([v[i] for v in c]) for i in range(len(c[0]))]
    return sum(euclidean_distance(mean_c, v) ** 2 for v in c)


if __name__ == "__main__":
    dataset: list[list[float]] = [[1.0, 1.0], [1.0, 2.0], [2.0, 1.0], [10.0, 1.0], [10.0, 2.0], [11.0, 1.0]]
    k = 6
    # Bisecting K-means testing
    diana = BisectingKmeans(k).fit(dataset)
    print(diana.cluster_tree_.get_clusters())
    print(diana.labels_)
    # diana.cluster_tree.plot()
    print(diana.predict(dataset))



from __future__ import annotations

import math
import statistics

from collections import Counter
from dataclasses import dataclass, field
from typing import Any, Optional


@dataclass
class KNN:
    """
    K-Nearest Neighbors classification algorithm implementation.

    This class implements the K-Nearest Neighbors algorithm for classification tasks.
    It supports optional standardization of the input data.

    Examples:
        >>> dataset = [[1.0, 2.0], [2.0, 3.0], [3.0, 4.0], [4.0, 5.0]]
        >>> labels = ['A', 'A', 'B', 'B']
        >>> knn = KNN(k=3, std_transform=True).fit(dataset, labels)
        >>> knn.predict([2.5, 3.5])
        'A'

    Attributes:
        k: The number of nearest neighbors to consider for classification.
        std_transform: Whether to standardize the input data (default: True).
        dataset_: The fitted dataset (standardized if std_transform is True).
        labels_: The labels corresponding to the fitted dataset.
        standardizationer_: The Standardizationer instance if std_transform is True.


    References:
        1. Li Hang
        2. Tan
        3. Zhou Zhihua
        4. Murphy
        5. Harrington
    """

    k: int
    std_transform: bool = True
    dataset_: Optional[list[list[float]]] = None
    labels_: Optional[list[Any]] = None
    standardizationer_: Optional[Standardizationer] = None

    def fit(self, dataset: list[list[float]], labels: list[Any]) -> KNN:
        """
        Fit the KNN model to the given dataset and labels.

        Args:
            dataset: The input dataset to fit the model to.
            labels: The labels corresponding to the input dataset.

        Returns:
            The fitted KNN instance.
        """
        self.dataset_ = dataset
        self.labels_ = labels
        if self.std_transform:
            self.standardizationer_ = Standardizationer()
            self.dataset_ = self.standardizationer_.fit_transform(self.dataset_)
        return self

    def predict(self, x: list[float]) -> Any:
        """
        Predict the label of the input data.

        Args:
            x: The input data to predict.

        Returns:
            The predicted label.

        Raises:
            ValueError: If the model is not fitted yet.
        """
        if self.dataset_ is None or self.labels_ is None:
            raise ValueError("The model is not fitted yet!")

        if self.std_transform:
            if self.standardizationer_ is None:
                raise ValueError("Cannot find the standardization!")
            x = self.standardizationer_.transform([x])[0]
        distances = [self._calculate_distance(x, point) for point in self.dataset_]
        # get k-nearest neighbors' label
        k_nearest_labels = [label for _, label in sorted(zip(distances, self.labels_), key=lambda x: x[0])][:: self.k]
        label = Counter(k_nearest_labels).most_common(1)[0][0]
        return label

    @staticmethod
    def _calculate_distance(x: list[float], y: list[float]) -> float:
        """
        Calculate the Euclidean distance between two points using a numerically stable method.

        This implementation avoids overflow by using the two-pass algorithm.
        """
        assert len(x) == len(y), f"{x} and {y} have different length!"

        # First pass: find the maximum absolute difference
        max_diff = max(abs(xi - yi) for xi, yi in zip(x, y))

        if math.isclose(max_diff, 0, abs_tol=1e-9):
            return 0.0  # All elements are identical

        # Second pass: calculate the normalized sum of squares
        sum_squares = sum(((xi - yi) / max_diff) ** 2 for xi, yi in zip(x, y))

        return max_diff * math.sqrt(sum_squares)


@dataclass
class Standardizationer:
    """
    A class for standardizing numerical datasets.

    Provides methods to fit a standardization model to a dataset,
    transform datasets using the fitted model, and perform both operations
    in a single step.
    """

    _means: list[float] = field(default_factory=list)
    _stds: list[float] = field(default_factory=list)
    _dimension: Optional[int] = None

    def fit(self, dataset: list[list[float]]) -> Standardizationer:
        """
        Fit the standardization model to the given dataset.

        Args:
            dataset: The input dataset to fit the model to.

        Returns:
            The fitted Standardizationer instance.

        Raises:
            ValueError: If the dataset has inconsistent dimensions.
        """
        self._dimension = self._get_dataset_dimension(dataset)
        self._means = self._dataset_column_means(dataset)
        self._stds = self._dataset_column_stds(dataset)
        return self

    def transform(self, dataset: list[list[float]]) -> list[list[float]]:
        """
        Transform the given dataset using the fitted standardization model.

        Args:
            dataset: The input dataset to transform.

        Returns:
            The standardized dataset.

        Raises:
            ValueError: If the model has not been fitted yet.
        """
        if self._dimension is None:
            raise ValueError("The model is not fitted yet!")
        return self.standardization(dataset)

    def fit_transform(self, dataset: list[list[float]]) -> list[list[float]]:
        """
        Fit the standardization model to the dataset and transform it in one step.

        Args:
            dataset: The input dataset to fit and transform.

        Returns:
            The standardized dataset.
        """
        self.fit(dataset)
        return self.transform(dataset)

    def standardization(self, dataset: list[list[float]]) -> list[list[float]]:
        """
        Standardize the given numerical dataset.
        The standardization is performed by subtracting the mean and dividing by the standard deviation for each feature.
        When the standard deviation is 0, all the values in the column are the same,
        here we set std to 1 to make every value in the column become 0 and avoid division by zero.

        Args:
            dataset: The input dataset to standardize.

        Returns:
            The standardized dataset.

        Raises:
            ValueError: If the model has not been fitted yet.
        """
        if self._dimension is None:
            raise ValueError("The model is not fitted yet!")
        for j, column in enumerate(zip(*dataset)):
            mean, std = self._means[j], self._stds[j]
            # ref: https://github.com/scikit-learn/scikit-learn/blob/7389dbac82d362f296dc2746f10e43ffa1615660/sklearn/preprocessing/data.py#L70
            if math.isclose(std, 0, abs_tol=1e-9):
                std = 1
            for i, value in enumerate(column):
                dataset[i][j] = (value - mean) / std
        return dataset

    @staticmethod
    def _get_dataset_dimension(dataset: list[list[float]]) -> int:
        dimension = len(dataset[0])
        if not all([len(row) == dimension for row in dataset]):
            raise ValueError("All rows must have the same number of columns")
        return dimension

    @staticmethod
    def _dataset_column_means(dataset: list[list[float]]) -> list[float]:
        """
        Calculate vectors mean
        """
        return [statistics.mean(column) for column in zip(*dataset)]

    @staticmethod
    def _dataset_column_stds(dataset: list[list[float]]) -> list[float]:
        """
        Calculate vectors(every column) standard variance
        """
        return [statistics.stdev(column) for column in zip(*dataset)]


if __name__ == "__main__":
    dataset: list[list[float]] = [[0.0], [1], [2], [3]]
    labels: list[Any] = [0, 0, 1, 1]
    knn = KNN(3).fit(dataset, labels)
    print(knn.labels_)
    print(knn.predict([1.1]))



from __future__ import annotations

import abc
import logging
import math

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Callable, List, Optional, Type

logger = logging.getLogger(__name__)


@dataclass
class BaseWeakLeaner:
    @abc.abstractmethod
    def fit(self, dataset: list[list[float]], weights: list[float], labels: list[int]) -> BaseWeakLeaner:
        raise NotImplementedError

    @abc.abstractmethod
    def predict(self, x: list[float]) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def get_predict_labels(self) -> list[int]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_error_rate(self) -> float:
        raise NotImplementedError


@dataclass
class AdaBoost:
    """
    The implementation of AdaBoost algorithm.

    Examples:
        >>> from toyml.ensemble.adaboost import AdaBoost, OneDimensionClassifier
        >>> dataset = [[0], [1], [2], [3], [4], [5], [6], [7], [8], [9]]
        >>> labels = [1, 1, 1, -1, -1, -1, 1, 1, 1, -1]
        >>> ada = AdaBoost(weak_learner=OneDimensionClassifier, n_weak_learner=3).fit(dataset, labels)
        >>> print(f"Training dataset error rate: {ada.training_error_rate_}")
        Training dataset error rate: 0.0
        >>> test_sample = [1.5]
        >>> print(f"The label of {test_sample} is {ada.predict(test_sample)}")
        The label of [1.5] is 1

    References:
        1. Li Hang
        2. Zhou Zhihua
    """

    weak_learner: Type[BaseWeakLeaner]
    """
    The weak learner to be used in the AdaBoost algorithm.
    """
    n_weak_learner: int = 5
    """
    The number of weak learners to be used in the AdaBoost algorithm.
    """
    predict_labels_: Optional[list[int]] = None
    """
    The prediction labels of the training dataset.
    """
    training_error_rate_: Optional[float] = None
    """
    The error rate of the training dataset.
    """

    _n: int = -1
    """
    The number of samples in the training dataset.
    """
    _labels: list[int] = field(default_factory=list)
    """
    The labels of the training dataset.
    """
    _weights: list[float] = field(default_factory=list)
    """
    The weights of samples in the training dataset.
    """
    _base_clf_labels: list[list[int]] = field(default_factory=list)
    """
    The prediction labels of the base classifiers.
    """
    _weak_learner_predicts: List[Callable[..., Any]] = field(default_factory=list)
    """
    The prediction functions of the weak learners.
    """
    _alphas: list[float] = field(default_factory=list)
    """
    The alpha values of the weak learners.
    """

    def fit(
        self,
        dataset: list[list[float]],
        labels: list[int],
    ) -> AdaBoost:
        """
        Fit the AdaBoost model.
        """
        self._labels = labels
        # for model training(Gm)
        self._n = len(labels)
        self._weights: Any = [1.0 / self._n] * self._n
        # we use -2 to initialize the class which can handle cases
        # such as multi-classes(0, 1, 2, ...) and binary classes(-1, 1)
        self._base_clf_labels = [[-2] * self._n for _ in range(self.n_weak_learner)]
        # base clf models
        self._weak_learner_predicts: List[Callable[..., int]] = []
        self._alphas = [0.0] * self.n_weak_learner

        for m in range(self.n_weak_learner):
            model = self.weak_learner().fit(dataset, self._weights, self._labels)
            self._base_clf_labels[m] = model.get_predict_labels()
            self._weak_learner_predicts.append(model.predict)
            error_rate = model.get_error_rate()
            # Warning when the error rate is too large
            if error_rate > 0.5:
                logger.warning(f"Weak learner error rate = {error_rate} < 0.5")
            alpha = 0.5 * math.log((1 - error_rate) / error_rate)
            self._alphas[m] = alpha
            # update the weights
            weights = [0.0] * self._n
            for i in range(self._n):
                weights[i] = self._weights[i] * math.exp(-alpha * self._labels[i] * self._base_clf_labels[m][i])
            self._weights = [weight / sum(weights) for weight in weights]
        # collect training dataset result
        self.predict_labels_ = [self.predict(x) for x in dataset]
        self.training_error_rate_ = sum(self.predict_labels_[i] != self._labels[i] for i in range(self._n)) / self._n
        return self

    def predict(self, x: list[float]) -> int:
        """
        Predict the label of the input sample.
        """
        ensemble_predict = 0
        for m in range(self.n_weak_learner):
            model_predict = self._weak_learner_predicts[m]
            ensemble_predict += self._alphas[m] * model_predict(x)
        if ensemble_predict >= 0:
            return 1
        else:
            return -1


@dataclass
class OneDimensionClassifier(BaseWeakLeaner):
    """
    Binary classifier with one dimension feature.

    Ref: Li Hang, 1 ed, E8.1.3
    """

    class SignMode(str, Enum):
        POS_NEG = "POS_NEG"
        NEG_POS = "NEG_POS"

    _sign_mode: SignMode = SignMode.POS_NEG
    _best_cut: float = math.inf
    error_rate_: float = math.inf
    predict_labels_: Optional[list[int]] = None

    def fit(
        self,
        dataset: list[list[float]],
        weights: list[float],
        labels: list[int],
    ) -> OneDimensionClassifier:
        """
        Fit the one-dimension classifier.
        """
        # search for the best cut point
        sign_mode, best_cut, best_error_rate = self.get_best_cut(dataset, weights, labels)
        self.error_rate_ = best_error_rate
        self._best_cut = best_cut
        self._sign_mode = sign_mode
        # get labels
        self.predict_labels_ = [0] * len(labels)
        for i, x in enumerate(dataset):
            self.predict_labels_[i] = self.predict(x)
        return self

    def predict(self, x: list[float]) -> int:
        """
        Predict the label of the input sample.
        """
        if self._best_cut is None:
            raise ValueError("The model is not fitted yet!")
        if self._sign_mode == "POS_NEG":
            if x[0] <= self._best_cut:
                return 1
            else:
                return -1
        if x[0] <= self._best_cut:
            return -1
        else:
            return 1

    def get_error_rate(self) -> float:
        """
        Get the error rate of the training dataset.
        """
        if self.error_rate_ is None:
            raise ValueError("The model is not fitted yet!")
        return self.error_rate_

    def get_predict_labels(self) -> list[int]:
        """
        Get the prediction labels of the training dataset.
        """
        if self.predict_labels_ is None:
            raise ValueError("The model is not fitted yet!")
        return self.predict_labels_

    @staticmethod
    def _get_candidate_cuts(points: list[float]) -> list[float]:
        """
        Get the candidate cuts of the training dataset.
        """
        min_x_int = math.floor(min(points))
        max_x_int = math.ceil(max(points))
        return [i + 0.5 for i in range(min_x_int, max_x_int)]

    def get_best_cut(
        self, dataset: list[list[float]], weights: list[float], labels: list[int]
    ) -> tuple[SignMode, float, float]:
        """
        Get the best cut of the training dataset.
        """
        points = [x[0] for x in dataset]
        candidate_cuts = self._get_candidate_cuts(points)
        # (func_mode, cut, error_rate)
        candidate_cuts_result = []
        for cut in candidate_cuts:
            pos_neg_error_rate = self._get_cut_error_rate(cut, points, weights, labels, self.SignMode.POS_NEG)
            neg_pos_error_rate = self._get_cut_error_rate(cut, points, weights, labels, self.SignMode.NEG_POS)
            candidate_cuts_result.extend(
                [(self.SignMode.POS_NEG, cut, pos_neg_error_rate), (self.SignMode.NEG_POS, cut, neg_pos_error_rate)]
            )

        # sorted by error rate
        best_cut_result = sorted(candidate_cuts_result, key=lambda x: x[2])[0]
        sign_mode, best_cut, best_error_rate = best_cut_result
        return sign_mode, best_cut, best_error_rate

    def _get_cut_error_rate(
        self,
        cut: float,
        points: list[float],
        weights: list[float],
        labels: list[int],
        sign_mode: SignMode,
    ) -> float:
        """
        Get the error rate of the training dataset.
        """
        if sign_mode == self.SignMode.POS_NEG:
            error_rate = sum(
                weights[i] for i, x in enumerate(points) if x <= cut and labels[i] != 1 or x > cut and labels[i] != -1
            )
        else:
            error_rate = sum(
                weights[i] for i, x in enumerate(points) if x <= cut and labels[i] != -1 or x > cut and labels[i] != 1
            )
        return error_rate


if __name__ == "__main__":
    dataset_demo: list[list[float]] = [[0], [1], [2], [3], [4], [5], [6], [7], [8], [9]]
    labels_demo: list[int] = [1, 1, 1, -1, -1, -1, 1, 1, 1, -1]
    ada = AdaBoost(weak_learner=OneDimensionClassifier, n_weak_learner=3).fit(dataset_demo, labels_demo)
    print(f"Training dataset prediction labels: {ada.predict_labels_}")
    print(f"Training dataset error rate: {ada.training_error_rate_}")
    test_sample = [1.5]
    print(f"The label of {test_sample} is {ada.predict(test_sample)}")



from __future__ import annotations

import math

from dataclasses import dataclass, field
from typing import Literal, Optional, Tuple


@dataclass
class AGNES:
    """
    Agglomerative clustering algorithm (Bottom-up Hierarchical Clustering)

    Examples:
        >>> from toyml.clustering import AGNES
        >>> dataset = [[1, 0], [1, 1], [1, 2], [10, 0], [10, 1], [10, 2]]
        >>> agnes = AGNES(n_cluster=2).fit(dataset)
        >>> print(agnes.labels_)
        [0, 0, 0, 1, 1, 1]

        >>> # Using fit_predict method
        >>> labels = agnes.fit_predict(dataset)
        >>> print(labels)
        [0, 0, 0, 1, 1, 1]

        >>> # Using different linkage methods
        >>> agnes = AGNES(n_cluster=2, linkage="complete").fit(dataset)
        >>> print(agnes.labels_)
        [0, 0, 0, 1, 1, 1]

        >>> # Plotting dendrogram
        >>> agnes = AGNES(n_cluster=1).fit(dataset)  # doctest: +SKIP
        >>> agnes.plot_dendrogram(show=True)  # doctest: +SKIP

    Tip: The AGNES Dendrogram Plot
        ![AGNES Dendrogram](../../images/agnes_dendrogram.png)

    Tip: References
        1. Zhou Zhihua
        2. Tan
    """

    n_cluster: int
    """The number of clusters, specified by user."""
    linkage: Literal["single", "complete", "average"] = "single"
    """The linkage method to use."""
    distance_metric: Literal["euclidean"] = "euclidean"
    """The distance metric to use.(For now we only support euclidean)."""
    distance_matrix_: list[list[float]] = field(default_factory=list)
    """The distance matrix."""
    clusters_: list[ClusterTree] = field(default_factory=list)
    """The clusters."""
    labels_: list[int] = field(default_factory=list)
    """The labels of each sample."""
    cluster_tree_: Optional[ClusterTree] = None
    linkage_matrix: list[list[float]] = field(default_factory=list)
    _cluster_index: int = 0

    def __post_init__(self) -> None:
        # check the linkage method
        if self.linkage not in ["single", "complete", "average"]:
            raise ValueError(
                f"Invalid linkage method: {self.linkage}, should be one of ['single', 'complete', 'average']"
            )
        # check the number of clusters
        if self.n_cluster < 1:
            raise ValueError("The number of clusters should be at least 1")

    def fit(self, dataset: list[list[float]]) -> AGNES:
        """
        Fit the model.
        """
        self._validate(dataset)
        n = len(dataset)
        self.clusters_ = [ClusterTree(cluster_index=i, sample_indices=[i]) for i in range(n)]
        self._cluster_index = n
        self.distance_matrix_ = self._get_init_distance_matrix(dataset)
        while len(self.clusters_) > self.n_cluster:
            (i, j), cluster_ij_distance = self._get_closest_clusters()
            # merge cluster_i and cluster_j
            self._merge_clusters(i, j, cluster_ij_distance)
            # update distance matrix
            self._update_distance_matrix(dataset, i, j)
        # build cluster_tree_
        self.cluster_tree_ = self._build_cluster_tree(n)
        # assign dataset labels
        self._get_labels(len(dataset))
        return self

    def fit_predict(self, dataset: list[list[float]]) -> list[int]:
        """
        Fit the model and return the labels of each sample.
        """
        self.fit(dataset)
        return self.labels_

    def _validate(self, dataset: list[list[float]]) -> None:
        """
        Validate the dataset.
        """
        n = len(dataset)
        # check the number of clusters
        if self.n_cluster > n:
            raise ValueError(
                "The number of clusters should be less than the number of data points,"
                "but got n_cluster={self.n_cluster} and n={n}"
            )
        # check the dataset rows
        if any(len(row) != len(dataset[0]) for row in dataset):
            raise ValueError("All samples should have the same dimension")

    def _get_clusters_distance(
        self,
        dataset: list[list[float]],
        cluster1: ClusterTree,
        cluster2: ClusterTree,
    ) -> float:
        """
        Get the distance between clusters c1 and c2 using the specified linkage method.
        """
        distances = [
            self._get_distance(dataset[i], dataset[j]) for i in cluster1.sample_indices for j in cluster2.sample_indices
        ]

        if self.linkage == "single":
            return min(distances)
        elif self.linkage == "complete":
            return max(distances)
        elif self.linkage == "average":
            return sum(distances) / len(distances)

    def _get_init_distance_matrix(self, dataset: list[list[float]]) -> list[list[float]]:
        """
        Gte initial distance matrix from sample points
        """
        n = len(dataset)
        distance_matrix = [[0.0 for _ in range(n)] for _ in range(n)]
        for i, cluster_i in enumerate(self.clusters_):
            for j, cluster_j in enumerate(self.clusters_):
                if j >= i:
                    distance_matrix[i][j] = self._get_clusters_distance(dataset, cluster_i, cluster_j)
                    distance_matrix[j][i] = distance_matrix[i][j]
        return distance_matrix

    def _get_closest_clusters(self) -> Tuple[Tuple[int, int], float]:
        """
        Search the distance matrix to get the closest clusters.
        """
        min_dist = math.inf
        closest_clusters = (0, 0)
        for i in range(len(self.distance_matrix_) - 1):
            for j in range(i + 1, len(self.distance_matrix_)):
                if self.distance_matrix_[i][j] < min_dist:
                    min_dist = self.distance_matrix_[i][j]
                    closest_clusters = (i, j)
        return closest_clusters, min_dist

    def _build_cluster_tree(self, n: int) -> ClusterTree:
        if len(self.clusters_) != 1:
            cluster_tree = ClusterTree(cluster_index=-1, sample_indices=list(range(n)))
            for cluster in self.clusters_:
                cluster_tree.add_child(cluster)
        else:
            cluster_tree = self.clusters_[0]
        return cluster_tree

    def _merge_clusters(self, i: int, j: int, cluster_ij_distance: float) -> None:
        """
        Merge two clusters to a new cluster.

        Args:
            i: the first indices of the clusters to merge
            j: the second indices of the clusters to merge
            cluster_ij_distance: the distance between the two clusters
        """
        cluster_i, cluster_j = self.clusters_[i], self.clusters_[j]
        # build new parent cluster
        parent_cluster = ClusterTree(cluster_index=self._cluster_index)
        self._cluster_index += 1
        # sort the sample indices for convenience
        parent_cluster.sample_indices = sorted(cluster_i.sample_indices + cluster_j.sample_indices)
        parent_cluster.add_child(cluster_i)
        parent_cluster.add_child(cluster_j)
        # parent_cluster.parent = self.cluster_tree_
        parent_cluster.children_cluster_distance = cluster_ij_distance
        # replace cluster1 with parent_cluster in clusters_
        self.clusters_[i] = parent_cluster
        self.clusters_.pop(j)

        # linkage matrix
        self.linkage_matrix.append(
            [cluster_i.cluster_index, cluster_j.cluster_index, cluster_ij_distance, len(parent_cluster.sample_indices)]
        )

    def _update_distance_matrix(self, dataset: list[list[float]], i: int, j: int) -> None:
        """
        Update the distance matrix after merging two clusters.
        """
        self.distance_matrix_.pop(j)
        # remove jth column
        for raw in self.distance_matrix_:
            raw.pop(j)
        # calc new dist
        for j in range(len(self.clusters_)):
            self.distance_matrix_[i][j] = self._get_clusters_distance(dataset, self.clusters_[i], self.clusters_[j])
            self.distance_matrix_[j][i] = self.distance_matrix_[i][j]

    def _get_labels(self, n: int) -> None:
        self.labels_ = [-1] * n
        for cluster_label, cluster in enumerate(self.clusters_):
            for sample_index in cluster.sample_indices:
                self.labels_[sample_index] = cluster_label

    def _get_distance(self, x: list[float], y: list[float]) -> float:
        assert len(x) == len(y), f"{x} and {y} have different length!"
        if self.distance_metric == "euclidean":
            return math.sqrt(sum(pow(x[i] - y[i], 2) for i in range(len(x))))
        else:
            raise ValueError(f"Distance metric {self.distance_metric} not supported!")

    def plot_dendrogram(
        self,
        figure_name: str = "agnes_dendrogram.png",
        show: bool = False,
    ) -> None:
        """
        Plot the dendrogram of the clustering result.

        This method visualizes the hierarchical structure of the clustering
        using a dendrogram. It requires the number of clusters to be set to 1
        during initialization.

        Args:
            figure_name: The filename for saving the plot.
                               Defaults to "agnes_dendrogram.png".
            show: If True, displays the plot. Defaults to False.

        Raises:
            ValueError: If the number of clusters is not 1.

        Note:
            This method requires matplotlib and scipy to be installed.
        """
        import matplotlib.pyplot as plt
        import numpy as np

        from scipy.cluster.hierarchy import dendrogram

        if self.n_cluster != 1:
            raise ValueError("The number of clusters should be 1 to plot dendrogram")
        # Plot the dendrogram
        plt.figure(figsize=(10, 7))
        dendrogram(np.array(self.linkage_matrix))
        plt.title("AGNES Dendrogram")
        plt.xlabel("Sample Index")
        plt.ylabel("Distance")
        plt.savefig(f"{figure_name}", dpi=300, bbox_inches="tight")
        if show:
            plt.show()


@dataclass
class ClusterTree:
    """
    Represents a node in the hierarchical clustering tree.

    Each node is a cluster containing sample indices.
    Leaf nodes represent individual samples, while internal nodes
    represent merged clusters. The root node contains all samples.
    """

    cluster_index: int
    parent: Optional[ClusterTree] = None
    """Parent node."""
    children: list[ClusterTree] = field(default_factory=list)
    """Children nodes."""
    sample_indices: list[int] = field(default_factory=list)
    """The cluster: dataset sample indices."""
    children_cluster_distance: Optional[float] = None

    def add_child(self, child: ClusterTree) -> None:
        child.parent = self
        self.children.append(child)

    def __repr__(self) -> str:
        return f"CT({self.cluster_index}): {self.sample_indices}"


if __name__ == "__main__":
    dataset: list[list[float]] = [[1.0, 2], [1, 5], [1, 0], [10, 2], [10, 5], [10, 0]]
    n_cluster: int = 2
    # fit
    agnes = AGNES(n_cluster).fit(dataset)
    for i in range(n_cluster):
        print(f"Cluster[{i}]: {agnes.clusters_[i].sample_indices}")
    # fit_predict
    y_pred = AGNES(n_cluster).fit_predict(dataset)
    print("Sample labels: ", y_pred)
    # Plot the dendrogram
    agnes = AGNES(n_cluster=1).fit(dataset)
    agnes.plot_dendrogram()



from __future__ import annotations

import logging
import math
import random

from dataclasses import dataclass, field
from typing import Literal, Optional

logger = logging.getLogger(__name__)


@dataclass
class Kmeans:
    """
    K-means algorithm (with Kmeans++ initialization as option).

    Examples:
        >>> from toyml.clustering import Kmeans
        >>> dataset = [[1.0, 2.0], [1.0, 4.0], [1.0, 0.0], [10.0, 2.0], [10.0, 4.0], [11.0, 0.0]]
        >>> kmeans = Kmeans(k=2, random_seed=42).fit(dataset)
        >>> kmeans.clusters_
        {0: [3, 4, 5], 1: [0, 1, 2]}
        >>> kmeans.centroids_
        {0: [10.333333333333334, 2.0], 1: [1.0, 2.0]}
        >>> kmeans.labels_
        [1, 1, 1, 0, 0, 0]
        >>> kmeans.predict([0, 1])
        1
        >>> kmeans.iter_
        2

    There is a `fit_predict` method that can be used to fit and predict.

    Examples:
        >>> from toyml.clustering import Kmeans
        >>> dataset = [[1, 0], [1, 1], [1, 2], [10, 0], [10, 1], [10, 2]]
        >>> Kmeans(k=2, random_seed=42).fit_predict(dataset)
        [1, 1, 1, 0, 0, 0]

    Tip: References
        1. Zhou Zhihua
        2. Murphy

    Note:
        Here we just implement the naive K-means algorithm.

    See Also:
      - Bisecting K-means algorithm: [toyml.clustering.bisect_kmeans][kmeans-bisect]
    """

    k: int
    """The number of clusters, specified by user."""
    max_iter: int = 500
    """The number of iterations the algorithm will run for if it does not converge before that."""
    tol: float = 1e-5
    """The tolerance for convergence."""
    centroids_init_method: Literal["random", "kmeans++"] = "random"
    """The method to initialize the centroids."""
    random_seed: Optional[int] = None
    """The random seed used to initialize the centroids."""
    distance_metric: Literal["euclidean"] = "euclidean"
    """The distance metric to use.(For now we only support euclidean)."""
    iter_: int = 0
    clusters_: dict[int, list[int]] = field(default_factory=dict)
    """The clusters of the dataset."""
    centroids_: dict[int, list[float]] = field(default_factory=dict)
    """The centroids of the clusters."""
    labels_: list[int] = field(default_factory=list)
    """The cluster labels of the dataset."""

    def __post_init__(self) -> None:
        random.seed(self.random_seed)

    def fit(self, dataset: list[list[float]]) -> "Kmeans":
        """
        Fit the dataset with K-means algorithm.

        Args:
            dataset: the set of data points for clustering

        Returns:
            self.
        """
        self.centroids_ = self._get_initial_centroids(dataset)
        for _ in range(self.max_iter):
            self.iter_ += 1
            prev_centroids = self.centroids_
            self._iter_step(dataset)
            if self._is_converged(prev_centroids):
                break
        return self

    def _iter_step(self, dataset: list[list[float]]) -> None:
        """
        Can be used to control the fitting process step by step.
        """
        self.clusters_ = self._get_clusters(dataset)
        self.centroids_ = self._get_centroids(dataset)
        self.labels_ = self._predict_dataset_labels(dataset)

    def fit_predict(self, dataset: list[list[float]]) -> list[int]:
        """
        Fit and predict the cluster label of the dataset.

        Args:
            dataset: the set of data points for clustering

        Returns:
            Cluster labels of the dataset samples.
        """
        return self.fit(dataset).labels_

    def predict(self, point: list[float]) -> int:
        """
        Predict the label of the point.

        Args:
            point: The data point to predict.

        Returns:
            The label of the point.

        """
        if len(self.centroids_) == 0:
            raise ValueError("The model is not fitted yet")
        return self._get_point_centroid_label(point, self.centroids_)

    def _calc_distance(self, x: list[float], y: list[float]) -> float:
        assert len(x) == len(y), f"{x} and {y} have different length!"
        if self.distance_metric == "euclidean":
            return math.sqrt(sum(pow(x[i] - y[i], 2) for i in range(len(x))))
        else:
            raise ValueError(f"Distance metric {self.distance_metric} not supported!")

    def _predict_dataset_labels(self, dataset: list[list[float]]) -> list[int]:
        labels = [-1] * len(dataset)
        for cluster_label, cluster in self.clusters_.items():
            for data_point_index in cluster:
                labels[data_point_index] = cluster_label
        return labels

    def _get_initial_centroids(self, dataset: list[list[float]]) -> dict[int, list[float]]:
        """
        get initial centroids by a simple random selection
        """
        if self.centroids_init_method == "random":
            return self._get_initial_centroids_random(dataset)
        elif self.centroids_init_method == "kmeans++":
            return self._get_initial_centroids_kmeans_plus(dataset)
        else:
            raise ValueError(f"Invalid centroids initialization method: {self.centroids_init_method}")

    def _is_converged(self, prev_centroids: dict[int, list[float]]) -> bool:
        """
        Check if the centroids converged.

        Args:
            prev_centroids: previous centroids

        Returns:
            Whether the centroids converged.
        """
        # check every centroid
        for i, centroid in self.centroids_.items():
            prev_centroid = prev_centroids[i]
            # check every dimension
            for j in range(len(prev_centroid)):
                if abs(prev_centroid[j] - centroid[j]) > self.tol:
                    return False
        return True

    def _get_initial_centroids_random(self, dataset: list[list[float]]) -> dict[int, list[float]]:
        """
        Get initial centroids by a simple random selection.

        Args:
            dataset: The dataset for clustering

        Returns:
            The initial centroids
        """
        data_points = random.sample(dataset, self.k)
        centroids = {i: data_point for i, data_point in enumerate(data_points)}
        return centroids

    def _get_initial_centroids_kmeans_plus(self, dataset: list[list[float]]) -> dict[int, list[float]]:
        """
        Get initial centroids by k-means++ algorithm.

        Args:
            dataset: The dataset for clustering

        Returns:
            The initial centroids
        """
        self.centroids_ = dict()
        self.centroids_[0] = random.choice(dataset)
        for i in range(1, self.k):
            min_distances = [self._get_min_square_distance(point) for point in dataset]
            total_dist = sum(min_distances)
            weights = [dist / total_dist for dist in min_distances]
            self.centroids_[i] = random.choices(dataset, weights)[0]
        return self.centroids_

    def _get_min_square_distance(self, point: list[float]) -> float:
        """
        Get the minimum square distance from the point to current centroids.

        Args:
            point: The point to calculate the distance.

        Returns:
            The minimum square distance
        """
        return min(
            (self._calc_distance(point, centroid) ** 2 for centroid in self.centroids_.values() if centroid),
            default=math.inf,
        )

    def _get_point_centroid_label(self, point: list[float], centroids: dict[int, list[float]]) -> int:
        """
        Get the label of the centroid, which is closest to the point
        """
        distances = [(i, self._calc_distance(point, centroid)) for i, centroid in centroids.items()]
        return min(distances, key=lambda x: x[1])[0]

    def _get_clusters(self, dataset: list[list[float]]) -> dict[int, list[int]]:
        clusters: dict[int, list[int]] = {i: [] for i in range(self.k)}
        for i, point in enumerate(dataset):
            centroid_label = self._get_point_centroid_label(point, self.centroids_)
            clusters[centroid_label].append(i)
        return clusters

    def _get_centroids(self, dataset: list[list[float]]) -> dict[int, list[float]]:
        centroids: dict[int, list[float]] = {i: [] for i in range(self.k)}
        for cluster_i, cluster in self.clusters_.items():
            cluster_points = [dataset[i] for i in cluster]
            centroid = [sum(t) / len(cluster) for t in zip(*cluster_points)]
            centroids[cluster_i] = centroid
        return centroids
